@@ -325,6 +325,7 @@ class Error(Token):
 @dataclass(order=True)
 class Instrument(Token):
     name: str
+    transpose: int = 0
 
 
 ###############################################################################
@@ -749,3 +750,11 @@ class Triplet(Token):
     """
 
     start: bool
+
+
+###############################################################################
+
+
+@dataclass
+class Vibrato(Token):
+    start: bool
